@@ -124,7 +124,7 @@ def center_wcs(slit, waves, center_on_source=False, force_nypix=31, fix_slope=No
     return wcs_data, offset_to_source, metadata_tuple(slit)
 
 
-def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True, drizzle_params=DRIZZLE_PARAMS, master_bkg=None, waves=None, wave_sample=1, log_step=True, force_nypix=31, center_on_source=False, center_phase=-0.5, fix_slope=None, outlier_threshold=5, sn_threshold=3, bar_threshold=0.7, err_threshold=1000, bkg_offset=5, bkg_parity=[1,-1], mask_padded=False, show_drizzled=True, show_slits=True, imshow_kws=IMSHOW_KWS, **kwargs):
+def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True, drizzle_params=DRIZZLE_PARAMS, master_bkg=None, wave_arrays={}, wave_sample=1, log_step=True, force_nypix=31, center_on_source=False, center_phase=-0.5, fix_slope=None, outlier_threshold=5, sn_threshold=3, bar_threshold=0.7, err_threshold=1000, bkg_offset=5, bkg_parity=[1,-1], mask_padded=False, show_drizzled=True, show_slits=True, imshow_kws=IMSHOW_KWS, get_quick_data=False,  max_sn_threshold=20, reopen=True, **kwargs):
     """
     Implementing more direct drizzling of multiple 2D slitlets
     
@@ -153,8 +153,8 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
     master_bkg : array-like
         Master background to replace local background derived from the drizzled product
     
-    waves : array-like
-        Explicit arget wavelength array
+    wave_arrays : dict
+        Explicit target wavelength arrays with keys for `{grating}-{filter}` combinations
     
     wave_sample, log_step : float, bool
         If `waves` not specified, generate with `msaexp.utils.get_standard_wavelength_grid`
@@ -213,8 +213,8 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
         `SlitModel` objects for the input slitlets
     
     drz_data : dict
-        3D `sci` and `wht` arrays of the drizzled slitlets that were combined into the drizzled
-        stack
+        3D `sci` and `wht` arrays of the drizzled slitlets that were combined
+        into the drizzled stack
     
     """
     from gwcs import wcstools
@@ -226,22 +226,30 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
         
     ### Read the SlitModels
     gratings = {}
+    grating_files = {}
+    
     for file in files:
         slit = jwst.datamodels.SlitModel(file)
+        utils.update_slit_metadata(slit)
+        
         grating = slit.meta.instrument.grating.lower()
         filt = slit.meta.instrument.filter.lower()
         key = f'{grating}-{filt}'
         if key not in gratings:
             gratings[key] = []
+            grating_files[key] = []
     
         gratings[key].append(slit)
-    
+        grating_files[key].append(file)
+        
     if verbose:
         for g in gratings:
-            msg = f'msaexp.drizzle.drizzle_slitlets: id={id}  {g} N={len(gratings[g])}'
-            grizli.utils.log_comment(grizli.utils.LOGFILE, msg, verbose=verbose, 
+            msg = f'msaexp.drizzle.drizzle_slitlets: id={id}  {g}'
+            msg += f' N={len(gratings[g])}'
+            grizli.utils.log_comment(grizli.utils.LOGFILE,
+                                     msg, 
+                                     verbose=verbose, 
                                      show_date=False)
-            
     
     ### DQ mask
     for gr in gratings:
@@ -252,7 +260,8 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
 
             if mask_padded:
                 msg = f'Mask padded area of slitlets: {mask_padded*1:.1f}'
-                grizli.utils.log_comment(grizli.utils.LOGFILE, msg, verbose=verbose, 
+                grizli.utils.log_comment(grizli.utils.LOGFILE,
+                                         msg, verbose=verbose, 
                                          show_date=False)
                 
                 _wcs = slit.meta.wcs
@@ -283,11 +292,14 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
         slits = gratings[gr]
         
         ## Default wavelengths
-        if waves is None:
-            waves = utils.get_standard_wavelength_grid(grating,
+        if gr in wave_arrays:
+            waves = wave_arrays[gr]
+        else:
+            waves = utils.get_standard_wavelength_grid(gr.split('-')[0],
                                                        sample=wave_sample,
                                                        log_step=log_step)
         
+        # print(gr, len(waves), waves.min(), waves.max())
         ## Drizzle 2D spectra
         drz = None
         drz_ids = []
@@ -305,6 +317,8 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
             slit = slits[i]
             if 'background' in slit.source_name:
                 continue
+            elif slit.data.shape[1] < 50:
+                continue
             
             _center = center_wcs(slit,
                                  waves,
@@ -314,7 +328,11 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
                                  center_phase=center_phase)
             
             wcs_data, offset_to_source, wcs_meta = _center
-            to_ujy = 1.e12*slit.meta.photometry.pixelarea_steradians
+            try:
+                to_ujy = 1.e12*slit.meta.photometry.pixelarea_steradians
+            except TypeError:
+                to_ujy = 1.0
+                
             break
         
         if wcs_meta is None:
@@ -328,7 +346,10 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
                                  center_phase=center_phase)
                                  
             wcs_data, offset_to_source, wcs_meta = _center
-            to_ujy = 1.e12*slits[0].meta.photometry.pixelarea_steradians
+            try:
+                to_ujy = 1.e12*slits[0].meta.photometry.pixelarea_steradians
+            except TypeError:
+                to_ujy = 1.0
 
         ##################
         # Now do the drizzling
@@ -366,7 +387,8 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
             msg = f'msaexp.drizzle.drizzle_slitlets: '
             msg += f'{gr} {i:2} {slit.source_name:18} {slit.source_id:9}'
             msg += f' {slit.source_ypos:.3f} {files[i]} {slit.data.shape}'
-            grizli.utils.log_comment(grizli.utils.LOGFILE, msg, verbose=verbose, 
+            grizli.utils.log_comment(grizli.utils.LOGFILE, msg, 
+                                     verbose=verbose, 
                                      show_date=False)
     
             drz_ids.append(slit.source_name)
@@ -386,20 +408,31 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
                 wcs_data, offset_to_source, wcs_meta = _center
                 
                 msg = f'msaexp.drizzle.drizzle_slitlets: '
-                msg += f'Recenter on source ({metadata_tuple(slit)}) y={offset_to_source:.2f}'
-                grizli.utils.log_comment(grizli.utils.LOGFILE, msg, verbose=verbose, 
+                msg += f'Recenter on source ({metadata_tuple(slit)})'
+                msg += f' y={offset_to_source:.2f}'
+                grizli.utils.log_comment(grizli.utils.LOGFILE,
+                                         msg, verbose=verbose, 
                                          show_date=False)
                 
                 # Recalculate photometry scaling
-                to_ujy = 1.e12*slit.meta.photometry.pixelarea_steradians
+                try:
+                    to_ujy = 1.e12*slit.meta.photometry.pixelarea_steradians
+                except TypeError:
+                    to_ujy = 1.0
             
             # Slit metadata
             keys = utils.slit_metadata_to_header(slit, key=i+1, header=h)
             
             # Do the drizzling
-            _waves, _header, _drz = utils.drizzle_slits_2d([slit],
-                                                           build_data=wcs_data,
-                                                           drizzle_params=drizzle_params)
+            if reopen:
+                _slit = jwst.datamodels.SlitModel(grating_files[gr][i])
+                utils.update_slit_metadata(_slit)
+            else:
+                _slit = slit
+                
+            _waves, _header, _drz = utils.drizzle_slits_2d([_slit],
+                                           build_data=wcs_data,
+                                           drizzle_params=drizzle_params)
             
             to_ujy_list.append(to_ujy)
             
@@ -407,12 +440,17 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
                 drz = _drz
             else:
                 drz.extend(_drz)
-        
+            
+            slit.close()
+            if reopen:
+                _slit.close()
+            
         drz_ids = np.array(drz_ids)
         
         ## Are slitlets tagged as background?
-        is_bkg = np.array([d.startswith('background') for d in drz_ids])
-        is_bkg &= False # ignore, combine them all
+        #is_bkg = np.array([d.startswith('background') for d in drz_ids])
+        #is_bkg &= False # ignore, combine them all
+        is_bkg = np.zeros(len(drz_ids), dtype=bool)
         
         ############ Combined drizzled spectra
         
@@ -422,11 +460,14 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
         err = np.array([d.err*to_ujy_list[i]
                         for i, d in enumerate(drz)])
         
+        if get_quick_data:
+            return sci, err
+            
         with warnings.catch_warnings():
             warnings.simplefilter(action='ignore', category=RuntimeWarning)
-            scima = np.nanmax(sci, axis=0)
+            scimax = np.nanmax(sci, axis=0)
         
-        flagged = (sci >= scima)
+        flagged = (sci >= scimax) & (sci/err > max_sn_threshold)
         flagged |= (err <= 0) | (~np.isfinite(sci)) | (~np.isfinite(err))
         flagged |= sci == 0
         flagged |= ~np.isfinite(sci/err)
@@ -434,7 +475,8 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
         
         for i in range(len(drz_ids)):
             ei = err[i,:,:]
-            emask =  ei > err_threshold*np.median(ei[np.isfinite(ei) & (ei > 0)])
+            emi = np.isfinite(ei) & (ei > 0)
+            emask =  ei > err_threshold*np.median(ei[emi])
             flagged[i,:,:] |= emask
         
         ivar = 1./err**2
@@ -446,7 +488,6 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
             avg = np.nanmedian(sci, axis=0)
             
         avg_w = ivar.sum(axis=0)
-        print('xxx init ', flagged.sum(), sci.size, avg.size)
         
         ## Subsequent passes - weighted average with outlier rejection
         for _iter in range(3):
@@ -465,7 +506,8 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
             
             for i in range(len(drz_ids)):
                 ei = err[i,:,:]
-                emask =  ei > err_threshold*np.median(ei[np.isfinite(ei) & (ei > 0)])
+                emi = np.isfinite(ei) & (ei > 0)
+                emask =  ei > err_threshold*np.median(ei[emi])
                 flagged[i,:,:] |= emask
             
             ivar = 1./err**2
@@ -509,7 +551,8 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
         
         # Valid data along wavelength axis
         xvalid = np.isfinite(avg).sum(axis=0) > 0
-        xvalid &= nd.binary_erosion(nd.binary_dilation(xvalid, iterations=2), iterations=4)
+        xvalid &= nd.binary_erosion(nd.binary_dilation(xvalid, iterations=2), 
+                                    iterations=4)
                 
         # Build HDUList
         h['EXTNAME'] = 'SCI'
@@ -544,7 +587,8 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
             dfig = None
 
         if (show_slits):
-            sfig = show_drizzled_slits(slits, sci, ivar, hdul, imshow_kws=imshow_kws,
+            sfig = show_drizzled_slits(slits, sci, ivar, hdul, 
+                                       imshow_kws=imshow_kws,
                                        with_background=(show_slits > 1))
             if output is not None:
                 sfig.savefig(f'{output}-{id}-{gr}.slit2d.png')
@@ -567,6 +611,31 @@ def drizzle_slitlets(id, wildcard='*phot', files=None, output=None, verbose=True
 def show_drizzled_slits(slits, sci, ivar, hdul, figsize=FIGSIZE, imshow_kws=IMSHOW_KWS, with_background=False):
     """
     Make a figure showing drizzled slitlets
+    
+    Parameters
+    ----------
+    slits : list
+        List of slitlet objects
+    
+    sci : (N,NY,NX) array
+        Science array of drizzled slitlets
+    
+    ivar : (N,NY,NX) array
+        Science array if inverse variance weights
+    
+    hdul : `~astropy.io.fits.HDUList`
+        Drizzle-combined HDU
+    
+    figsize : tuple
+        Figure size
+    
+    imshow_kws : dict
+        Keywords passed to `~matplotlib.pyplot.imshow`
+    
+    Returns
+    -------
+    fig : Figure
+    
     """
     avg = hdul['SCI'].data
     xvalid = np.isfinite(avg).sum(axis=0) > 0
@@ -609,6 +678,23 @@ def show_drizzled_slits(slits, sci, ivar, hdul, figsize=FIGSIZE, imshow_kws=IMSH
 def show_drizzled_product(hdul, figsize=FIGSIZE, imshow_kws=IMSHOW_KWS):
     """
     Make a figure showing drizzled product
+    
+    Parameters
+    ----------
+    hdul : `~astropy.io.fits.HDUList`
+        Drizzle combined HDU
+    
+    figsize : tuple
+        Figure size
+    
+    imshow_kws : dict
+        kwargs for `~matplotlib.pyplot.imshow`
+    
+    Returns
+    -------
+    fig : Figure
+        Figure object
+    
     """
     
     avg = hdul['SCI'].data
@@ -651,13 +737,75 @@ def show_drizzled_product(hdul, figsize=FIGSIZE, imshow_kws=IMSHOW_KWS):
     return fig
 
 
-def make_optimal_extraction(waves, sci2d, wht2d, profile_slice=None,
-                            prf_center=None, prf_sigma=1.0, sigma_bounds=(0.5, 2.5), 
-                            center_limit=4,
-                            fit_prf=True, fix_center=False, fix_sigma=False, trim=0,
-                            bkg_offset=6, bkg_parity=[-1,1], verbose=True, **kwargs):
+def make_optimal_extraction(waves, sci2d, wht2d, profile_slice=None, prf_center=None, prf_sigma=1.0, sigma_bounds=(0.5, 2.5), center_limit=4, fit_prf=True, fix_center=False, fix_sigma=False, trim=0, bkg_offset=6, bkg_parity=[-1,1], verbose=True, **kwargs):
     """
     Optimal extraction from 2D arrays
+                            
+    Parameters
+    ----------
+    waves : 1D array
+        Wavelengths, microns
+    
+    sci2d : 2D array
+        Data array
+    
+    wht2d : 2D array
+        Inverse variance weight array
+    
+    profile_slice : tuple, slice
+        Slice along wavelength axis where to determine the cross-dispersion profile.  If a 
+        tuple of floats, interpret as wavelength limits in microns
+    
+    prf_center : float
+        Profile center, relative to the cross-dispersion center of the array.  If `None`, then
+        try to estimate it from the data
+    
+    prf_sigma : float
+        Width of the extraction profile in pixels
+    
+    sigma_bounds : (float, float)
+        Parameter bounds for `prf_sigma`
+    
+    center_limit : float
+        Maximum offset from `prf_center` allowed
+    
+    fit_prf : bool
+        Fit the profile.  If False, then just use the fixed values of `prf_center` and 
+        `prf_sigma`
+    
+    fix_center : bool
+        Fix the centering in the fit
+    
+    fix_sigma : bool
+        Fix the width in the fit
+    
+    bkg_offset, bkg_parity : int, list
+        Parameters for the local background determination (see 
+        `~msaexp.drizzle.drizzle_slitlets`).  The profile is "subtracted" in the
+        same way as the data.
+    
+    verbose : bool
+        Status messages
+    
+    kwargs : dict
+        Ignored keyword args
+    
+    Returns
+    -------
+    sci2d_out : array
+        Output 2D sci array
+    
+    wht2d_out : array
+        Output 2D wht array
+    
+    profile2d : array
+        2D optimal extraction profile
+    
+    spec : `~astropy.table.Table`
+        Optimally-extracted 1D spectrum
+    
+    prof_tab : `~astropy.table.Table`
+        Table of the collapsed 1D profile
     """
     import scipy.ndimage as nd
     import astropy.units as u
@@ -688,9 +836,9 @@ def make_optimal_extraction(waves, sci2d, wht2d, profile_slice=None,
             
         slice_limits = profile_slice.start, profile_slice.stop
         
-        pmask = ok & False
-        pmask[:,profile_slice] = True
-        ok &= ~pmask
+        pmask = ok & True
+        pmask[:,profile_slice] &= True
+        ok &= pmask
         
     else:
         prof1d = np.nansum(sci2d * wht2d, axis=1) / np.nansum(wht2d, axis=1)
@@ -720,8 +868,15 @@ def make_optimal_extraction(waves, sci2d, wht2d, profile_slice=None,
     if fit_type == 0:
         args = (waves, sci2d, wht_mask, prf_center, prf_sigma,
                 bkg_offset, bkg_parity, 3, 1, (verbose > 1))
+
         pnorm, pmodel = utils.objfun_prf([prf_center, prf_sigma], *args)
         profile2d = pmodel/pnorm
+        pmask = (profile2d > 0) & np.isfinite(profile2d)
+        profile2d[~pmask] = 0
+        
+        fit_center = prf_center
+        fit_sigma = prf_sigma
+        
     else:
         # Fit it
         if fix_sigma:
@@ -764,7 +919,7 @@ def make_optimal_extraction(waves, sci2d, wht2d, profile_slice=None,
         pfit1d = np.nansum((wht2d*profile2d*sci1d)[:,profile_slice], axis=1) 
         pfit1d /= np.nansum(wht2d[:,profile_slice], axis=1)
     else:
-        pfit1d = np.nansum(profile2d*sci1d * wht2d, axis=1) / np.nansum(wht2d, axis=1)
+        pfit1d = np.nansum(profile2d*sci1d*wht2d, axis=1) / np.nansum(wht2d, axis=1)
     
     if trim > 0:
         bad = nd.binary_dilation(wht1d <= 0, iterations=trim)
@@ -817,6 +972,31 @@ def make_optimal_extraction(waves, sci2d, wht2d, profile_slice=None,
 
 def extract_from_hdul(hdul, prf_center=None, master_bkg=None, verbose=True, **kwargs):
     """
+    Run 1D extraction on arrays from a combined dataset
+    
+    Parameters
+    ----------
+    hdul : `~astropy.io.fits.HDUList`
+        Output data from from `~msaexp.drizzle.drizzle_slitlets`
+    
+    prf_center : float, None
+        Initial profile center.  If not specified, get from the `'SRCYPIX'` keyword in 
+        `hdul['SCI'].header`
+    
+    master_bkg : array
+        Optional master background to use instead of `hdul['BKG'].data`
+    
+    verbose : bool
+        Printing status messages
+    
+    kwargs : dict
+        Keyword arguments passed to `msaexp.drizzle.make_optimal_extraction`
+    
+    Returns
+    -------
+    outhdu : `~astropy.io.fits.HDUList`
+        Modified HDU including 1D extraction
+    
     """
         
     if master_bkg is None:
